function [sMap,sTrain,Qe,Te] = somng_batchtrain(sMap, D, varargin)
%% Map definition values
[munits,dim]= size(sMap.codebook);
if isstruct(D),
    sMap.comp_norm=D.comp_norm;
    sMap.comp_names=D.comp_names;
    sTrain.data_name=D.name;
    D=D.data;
end
[dlen, ~]= size (D);

sTrain.type='SOM-NG_train';
sTrain.algorithm='supervised batch';

sMap.start=clock;
sMap.neigh='ranking2';
mask=sMap.mask;
%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
a=1;
while a<=length(varargin),
    if strcmp(varargin(a),'gamma'), a=a+1; gamma=varargin{a};end
    if strcmp(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmp(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmp(varargin(a),'sigma'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'radius'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'radius_ini'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'euclidean'), sMap.neigh = 'euclidean'; end
    if strcmp(varargin(a),'ranking'), sMap.neigh = 'ranking'; end
    if strcmp(varargin(a),'ranking2'), sMap.neigh = 'ranking2'; end
    if strcmp(varargin(a),'figure'), a=a+1; Hfig=varargin{a}; end
    
    a=a+1;
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=50; end
if ~(exist('tracking','var')), tracking=-1-trainlen/1e4; end
if ~(exist('Hfig','var'))&&(abs(tracking)>=2), Hfig=figure(); end

%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.001
    sigma(1:trainlen)=(munits/2)*(0.002/munits).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:trainlen)=(sigma)*(0.001/sigma).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:trainlen)=(sigma(1))*(sigma(2)/sigma(1)).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==trainlen
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error
sTrain.radius_ini=sigma(1);
sTrain.radius_fin=sigma(end);

%% Topology preservation constant gamma
if ~(exist('gamma','var'))
    % Default value. Constant value of 1
    gamma=ones(1,trainlen);
    sTrain.gamma=1;
elseif length(gamma)==1
    % Constant
    sTrain.gamma=gamma;
    gamma=gamma*ones(1,trainlen);
else
    sTrain.gamma=gamma;
    aux=gamma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, gamma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
%% Determine the ranks of the map units s(ri*,ri)
Ud=(som_unit_dists(sMap.topol)).^2;
s=zeros(munits);
for a=1:munits
    [d,inds]=sort(Ud(:,a));
    switch (sMap.neigh)
        case 'ranking', s(inds,a) = 0:munits-1;
        case 'ranking2', s(inds,a) = 0:munits-1;
            for b=2:munits
                if (d(b)==d(b-1)), s(inds(b),a)=s(inds(b-1),a); end
            end
        case 'euclidean', s(inds,a) = Ud(inds,a);
    end
end
sTrain.neigh=sMap.neigh;
qe=zeros(2,trainlen);
start=clock;
sTrain.trainlen=trainlen;
sTrain.start=datestr(now);
%% For each epoch
for t=1:trainlen
    %% Tracking
    if tracking||(nargout>2)
        [qe(1,t),qe(2,t)]=som_quality(sMap,D);
        qe(:,t)=real(qe(:,t));
    end
    if tracking,
        if exist('Hfig','var'),figure(Hfig); end
        trackplot(sMap.codebook,D,tracking,start,t,qe);
    end
    %% Determine the ranks of the prototipes k(v,wi)
    Hng=zeros(munits,dlen);
    d=zeros(munits,dlen);
    for a=1:dlen
        v=D(a,:);
        d(:,a)=(((v(ones(munits,1),1:dim)-sMap.codebook).^2).*mask(:,ones(munits,1))')*ones(dim,1);
        [~,inds]=sort(d(:,a));
        Hng(inds,a)=0:munits-1;
    end
    %% SOM kernel
    %    bmus=som_bmus(sMap,D);
    [bmus,~]=find(Hng==0);
    Hsom=s(:,bmus);
    Hsom=exp(-Hsom/sigma(t));
    Hng=exp(-(Hng)/(gamma(t)^2*sigma(t)));
    
    %% Updating prototipes
    H=Hsom.*Hng;
    num=H*D;
    den=H*ones(dlen,1);
    nonzero=find(den>eps);
    sMap.codebook(nonzero,:)=num(nonzero,:)./den(nonzero,ones(dim,1));
end %for each epoch
sTrain.end=datestr(now);
sMap.trainhist={sMap.trainhist; sTrain};
if nargout>2,
    Qe=qe(1,:);Te=qe(2,:);
end
return % function
