function [sMap,sTrain,Qe,Te] = somng_blocktrain(sMap, D, varargin)
%% Map definition values
[munits,dim]= size(sMap.codebook);
if isstruct(D) D=D.data; end
[dlen,~] = size (D);
sMap.start=clock;
sMap.neigh='ranking2';
mask=sMap.mask;
%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
a=1;
while a<=length(varargin),
    if strcmp(varargin(a),'gamma'), a=a+1; gamma=varargin{a};end
    if strcmp(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmp(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmp(varargin(a),'sigma'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'gaussian'), sMap.neigh = 'gaussian'; end
    if strcmp(varargin(a),'ranking'), sMap.neigh = 'ranking'; end
    if strcmp(varargin(a),'ranking2'), sMap.neigh = 'ranking2'; end
    if strcmp(varargin(a),'figure'), a=a+1; Hfig=varargin{a}; end
    if strcmp(varargin(a),'blocksize'), a=a+1; bloque=min(varargin{a},2000); end
    
    a=a+1;
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=50; end
if ~(exist('tracking','var')), tracking=-1-trainlen/1e4; end
if ~(exist('Hfig','var'))&&(abs(tracking)>=2), Hfig=figure(); end
if ~(exist('bloque','var')), bloque=ceil(dlen/10); end

%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.001
    sigma(1:trainlen)=(munits/2)*(0.002/munits).^((0:trainlen-1)/(trainlen));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:trainlen)=(sigma)*(0.001/sigma).^((0:trainlen-1)/trainlen);
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:trainlen)=(sigma(1))*(sigma(2)/sigma(1)).^((0:trainlen-1)/trainlen);
elseif length(sigma)==trainlen
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error

%% Topology preservation constant gamma
if ~(exist('gamma','var'))
    % Default value. Constant value of 1
    gamma=ones(1,trainlen);
elseif length(gamma)==1
    % Constant
    gamma=gamma*ones(1,trainlen);
else
    aux=gamma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, gamma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
%% Determine the ranks of the map units s(ri*,ri)
Ud=(som_unit_dists(sMap.topol)).^2;
s=zeros(munits);
for a=1:munits
    [d,inds]=sort(Ud(:,a));
    switch (sMap.neigh)
        case 'ranking', s(inds,a) = 0:munits-1;
        case 'ranking2', s(inds,a) = 0:munits-1;
            for b=2:munits
             if (d(b)==d(b-1)), s(inds(b),a)=s(inds(b-1),a); end
            end
        case 'gaussian', s(inds,a) = Ud(inds,a);
    end
end
qe=zeros(2,trainlen);
start=clock;
%% For each epoch
for t=1:trainlen
%% Tracking
  if tracking,
    [qe(1,t),qe(2,t)]=som_quality(sMap,D);
    qe(:,t)=real(qe(:,t));
    if abs(tracking)>=2, figure(Hfig); end
    trackplot(sMap.codebook,D,tracking,start,t,qe);
  end
%% Determine the ranks of the prototipes k(v,wi)
d=zeros(munits,min(bloque,dlen));
posicion=1;num=0;den=0;
while (posicion<dlen)
Hng=zeros(munits,min(bloque,dlen-posicion+1));
for a=posicion:min(posicion+bloque-1,dlen)
    v=D(a,:);
    d(:,a-posicion+1)=(((v(ones(munits,1),1:dim)-sMap.codebook).^2).*(mask(:,ones(munits,1))'))*ones(dim,1);
    [~,inds]=sort(d(:,a-posicion+1));
    Hng(inds,a-posicion+1)=0:munits-1;
end
    Hng=exp(-(Hng)/(gamma(t)^2*sigma(t)));
%% SOM kernel
    bmus=som_bmus(sMap,D(posicion:min(posicion+bloque-1,dlen),:));
    Hsom=s(:,bmus);
    Hsom=exp(-Hsom/sigma(t));
%% Updating prootipes
    H=Hsom.*Hng;
    num=num+H*D(posicion:min(posicion+bloque-1,dlen),:);
    den=den+H*ones(min(bloque,dlen-posicion+1),1);
    posicion=posicion+bloque;
end
    nonzero=find(den>eps);
    sMap.codebook(nonzero,:)=num(nonzero,:)./den(nonzero,ones(dim,1));
end %for each epoch
sTrain=[];
Qe=qe(1,:);Te=qe(2,:);
sMap.end=clock;
sMap.algorithm='SOM-NG block batch';
return % function
