function [sMap,sTrain,Qe,Te] = somng_seqtrain(sMap, D, varargin)
% Copyright (C) 2017. University of Oviedo. All rights reserved.
% Author: Ivn Machn-Gonzlez; Mario Crespo
%
% Please, if you use it, cite the following work:
% Ivn Machn-Gonzlez and Hilario Lpez-Garca, Feedforward Nonlinear Control Using Neural Gas Network, Complexity, vol. 2017, Article ID 3125073, 11 pages, 2017. doi:10.1155/2017/3125073

%% Map definition values
[munits,dim]= size(sMap.codebook);
if isstruct(D) D=D.data; end
[dlen, ~]= size (D);

sMap.start=clock;
sMap.neigh='ranking2';
mask=sMap.mask;
%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
a=1;
while a<=length(varargin),
    if strcmpi(varargin(a),'alpha'), a=a+1; alpha=varargin{a};end
    if strcmpi(varargin(a),'gamma'), a=a+1; gamma=varargin{a};end
    if strcmpi(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmpi(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmpi(varargin(a),'sigma'), a=a+1; sigma=varargin{a}; end
    if strcmpi(varargin(a),'radius'), a=a+1; sigma=varargin{a}; end
    if strcmpi(varargin(a),'radius_ini'), a=a+1; sigma(1)=varargin{a}(1); end
    if strcmpi(varargin(a),'radius_fin'), a=a+1; sigma(2)=varargin{a}(1); end
    if strcmpi(varargin(a),'euclidean'), sMap.neigh = 'euclidean'; end
    if strcmpi(varargin(a),'ranking'), sMap.neigh = 'ranking'; end
    if strcmpi(varargin(a),'ranking2'), sMap.neigh = 'ranking2'; end
    if strcmpi(varargin(a),'figure'), a=a+1; Hfig=varargin{a}; end
    
    a=a+1;
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=50; end
if ~(exist('tracking','var')), tracking=0; end
if ~(exist('Hfig','var'))&&(abs(tracking)>=2), Hfig=figure(); end


%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.001
    sigma(1:(trainlen*dlen))=(munits/2)*(0.002/munits).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:(trainlen*dlen))=(sigma)*(0.001/sigma).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:(trainlen*dlen))=(sigma(1))*(sigma(2)/sigma(1)).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(sigma)>=(trainlen*dlen)
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,(trainlen*dlen),length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error
%% Learning rate alpha
if ~(exist('alpha','var'))
    % Default value. From 0.5 to 0.001
    alpha(1:(trainlen*dlen))=0.5*(0.001/0.5).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(alpha)==1
    % If only a value is given it will be initial radius
    alpha(1:(trainlen*dlen))=(alpha)*(0.001/alpha).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(alpha)==2
    % If two values are given, they will be initial and final radius
    alpha(1:trainlen)=(alpha(1))*(alpha(2)/alpha(1)).^((0:(trainlen*dlen)-1)/((trainlen*dlen)-1));
elseif length(alpha)>=(trainlen*dlen)
    % Alpha is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=alpha;
    t=round(linspace(1,(trainlen*dlen),length(aux)));
    for a=1:length(aux)-1, alpha(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end

%% Topology preservation constant gamma
if ~(exist('gamma','var'))
    % Default value. Constant value of 1
    gamma=ones(1,trainlen*dlen);
elseif length(gamma)==1
    % Constant
    gamma=gamma*ones(1,trainlen*dlen);
else
    aux=gamma;
    t=round(linspace(1,trainlen*dlen,length(aux)));
    for a=1:length(aux)-1, gamma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
%% Determine the ranks of the map units s(ri*,ri)
Ud=(som_unit_dists(sMap.topol)).^2;
s=zeros(munits);
for a=1:munits
    [d,inds]=sort(Ud(:,a));
    switch (sMap.neigh)
        case 'ranking', s(inds,a) = 0:munits-1;
        case 'ranking2', s(inds,a) = 0:munits-1;
            for b=2:munits
                if (d(b)==d(b-1)), s(inds(b),a)=s(inds(b-1),a); end
            end
        case 'euclidean', s(inds,a) = Ud(inds,a);
    end
end

if tracking||(nargout>2), qe=zeros(2,trainlen); end
start=clock;
%% For each epoch
for t=1:trainlen
    [~,ord]=sort(rand(dlen,1));
    % Hng=zeros(munits,dlen);
    % d=zeros(munits,dlen);
    for a=1:dlen
        step=(t-1)*dlen+a;
        v=D(ord(a),:);
        d=(((v(ones(munits,1),1:dim)-sMap.codebook).^2).*mask(:,ones(munits,1))')*ones(dim,1);
        [~,inds]=sort(d);
        Hng(inds,1)=0:munits-1; %#ok<AGROW>
        Hsom=s(:,(Hng==0));
        Hng=exp(-Hng./(gamma(step).^2*sigma(step)));
        Hsom=exp(-Hsom/sigma(step));
        H=Hsom.*Hng;
        sMap.codebook=sMap.codebook+alpha(step)*(H*v-H(:,ones(1,dim)).*sMap.codebook);
    end
    %% Tracking
    if tracking||(nargout>2)
        [qe(1,t),qe(2,t)]=som_quality2(sMap,D);
        qe(:,t)=real(qe(:,t));
    end
    if tracking,
        if exist('Hfig','var'),figure(Hfig); end
        trackplot(sMap.codebook,D,tracking,start,t,qe);
    end    
end %for each epoch

sTrain=[];
if nargout>2,
Qe=qe(1,:);Te=qe(2,:);
end
sMap.end=clock;
sMap.algorithm='SOM-NG seq';
return % function
