function [sMap] = somngs_blocktrain(sMap, D, varargin)
if ~isstruct(sMap)
    Map=som_randinit(sMap,'rect','msize',[floor(sqrt(length(sMap))) floor(sqrt(length(sMap)))]);
    Map.codebook=sMap;
    Map.vectors=sMap;
    Map.values=sMap(:,1);
    sMap=Map;
    clear Map;
end
mask=sMap.mask;

if isstruct(D)
    sMap.comp_norm=D.comp_norm;
    sMap.comp_names=D.comp_names;
    sMap.name=D.name;
    if isfield (D,'value_norm'), sMap.value_norm=D.value_norm; end
    if isfield (D,'value_name'), sMap.value_name=D.value_name; end
    if isfield (D,'values'),
        F=D.values;
        a=1;
    else
        F=varargin{1};
        a=2;
    end
    D=D.data;
else
    F=varargin{1};
    a=2;
end

if ~isnumeric(F), error('Invalid output values or not provided'); end
if size(F,1)~=size(D,1), error('Output values must have same length as input vectors'); end
if size(sMap.codebook,2)~=size(D,2), error('Initial map and data must have the same number of variables (columns)'); end
if any(any(isnan(D))), error ('Data must not contain any NaN'); end
if any(any(isnan(F))), error ('Output values must not contain any NaN'); end

%% Map definition values
[munits,dim]= size(sMap.codebook);
dlen = length (D);

% Default option
sMap.neigh='ranking2';

%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
a=1;
while a<=length(varargin),
    %Valued variables
    if strcmp(varargin(a),'gamma'), a=a+1; gamma=varargin{a};end
    if strcmp(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmp(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmp(varargin(a),'sigma'), a=a+1; sigma=varargin{a};end
    if strcmp(varargin(a),'values'), a=a+1; sMap.values=varargin{a}; end
    if strcmp(varargin(a),'vectors'), a=a+1; sMap.vectors=varargin{a}; end
    if strcmp(varargin(a),'outputname'), a=a+1; sMap.value_name=varargin{a}; end
    if strcmp(varargin(a),'blocksize'), a=a+1; bloque=min(varargin{a},2000); end    
    
    % Setup options
    if strcmp(varargin(a),'gaussian'), sMap.neigh = 'gaussian'; end
    if strcmp(varargin(a),'ranking'), sMap.neigh = 'ranking'; end
    if strcmp(varargin(a),'ranking2'), sMap.neigh = 'ranking2'; end
    if strcmp(varargin(a),'Hng'), s=zeros(munits); end
    
    
    a=a+1;
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=50; end
if ~(exist('tracking','var')), tracking=0; end
if ~(exist('bloque','var')), bloque=ceil(dlen/20); end

%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.001
    sigma(1:trainlen)=(munits/2)*(0.002/munits).^((0:trainlen-1)/(trainlen));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:trainlen)=(sigma)*(0.001/sigma).^((0:trainlen-1)/trainlen);
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:trainlen)=(sigma(1))*(sigma(2)/sigma(1)).^((0:trainlen-1)/(trainlen));
elseif length(sigma)>=trainlen
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error

%% Topology preservation constant gamma
if ~(exist('gamma','var'))
    % Default value. Constant value of 1
    gamma=ones(1,trainlen);
elseif length(gamma)==1
    % Constant
    gamma=gamma*ones(1,trainlen);
else
    aux=gamma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, gamma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
%% Determine the ranks of the map units s(ri*,ri)
if ~(exist('s','var'))
Ud=(som_unit_dists(sMap.topol)).^2;
s=zeros(munits);
for a=1:munits

    switch (sMap.neigh)
        case 'rankingm',    [~,inds]=sort(Ud(:,a));  s(inds,a) = 0:munits-1;
        case 'ranking',
            for b=1:munits, s(b,a)=sum(Ud(:,a)<Ud(b,a)); end
        case 'euclidean', s(:,a) = Ud(:,a);
    end
end
end

trainstr.neigh=sMap.neigh;
if abs(tracking)>=1, qe=zeros(2,2*trainlen); end
start=clock;
trainstr.trainlen=trainlen;
%% For each epoch
trainstr.start=datestr(now);
for t=1:trainlen
    %% Determine the ranks of the prototipes k(v,wi)
    d=zeros(munits,min(bloque,dlen));
    posicion=1;num=0;den=0;numY=0;
    while (posicion<dlen)
        Hng=zeros(munits,min(bloque,dlen-posicion+1));
        for a=posicion:min(posicion+bloque-1,dlen)
            v=D(a,:);
            d(:,a-posicion+1)=(((v(ones(munits,1),1:dim)-sMap.codebook).^2).*mask(:,ones(munits,1))')*ones(dim,1);
            [~,inds]=sort(d(:,a-posicion+1));
            Hng(inds,a-posicion+1)=0:munits-1;
        end
        Hng=exp(-(Hng)/(gamma(t)^2*sigma(t)));
        %% SOM kernel
        bmus=som_bmus(sMap,D(posicion:min(posicion+bloque-1,dlen),:));
        Hsom=s(:,bmus);
        Hsom=exp(-Hsom/sigma(t));
        %% Updating prootipes
        H=Hsom.*Hng;
        num=num+H*D(posicion:min(posicion+bloque-1,dlen),:);
        numY=numY+H*F(posicion:min(posicion+bloque-1,dlen));
        den=den+H*ones(min(bloque,dlen-posicion+1),1);
        posicion=posicion+bloque;
    end
    nonzero=find(den>eps);
    sMap.codebook(nonzero,:)=num(nonzero,:)./den(nonzero,ones(dim,1));
    sMap.values(nonzero)=numY(nonzero)./den(nonzero);
  
    %% Tracking
    if abs(tracking)>=1,
        [qe(1,t),qe(2,t)]=som_quality(sMap,D);
        qe(:,t)=real(qe(:,t));
        trackplot([sMap.codebook sMap.values],[D F],tracking,start,t,qe);
    end

end %for each epoch
    

for t=1:trainlen
    posicion=1; numA=zeros(munits,dim); denA=zeros(munits,1);
    while (posicion<dlen)
        d=zeros(munits,min(bloque,dlen-posicion+1));
        Hng=d; k=Hng;
        for a=posicion:min(posicion+bloque-1,dlen)
            d(:,a-posicion+1)=(D(ones(munits,1)*a,1:dim)-sMap.codebook).^2*ones(dim,1);
            [~,inds]=sort(d(:,a-posicion+1));
            k(inds,a-posicion+1)=0:munits-1;
        end
        
        %% Determine the ranks of the prototipes k(v,wi)
        Hng=exp(-(k)/(gamma(t)^2*sigma(t)));
        %% SOM kernel
        bmus=som_bmus(sMap,D(posicion:min(posicion+bloque-1,dlen),:));
        Hsom=exp(-s/sigma(t));
        %% Updating prootipes
        P = sparse(bmus,1:length(bmus),1,munits,length(bmus));
        H=(Hsom*P).*Hng;
        
        for a=1:munits
            Dist=D(posicion:min(posicion+bloque-1,dlen),:)-sMap.codebook(ones(min(bloque,dlen-posicion+1),1)*a,:);
            E=F(posicion:min(posicion+bloque-1,dlen))-sMap.values(ones(min(bloque,dlen-posicion+1),1)*a,:)-((sMap.vectors(ones(min(bloque,dlen-posicion+1),1)*a,:).*Dist)*ones(dim,1));
            numA(a,:)=numA(a,:)+sum((H(ones(dim,1)*a,:)').*E(:,ones(dim,1)).*Dist);
            denA(a)=denA(a)+sum((H(a,:)').*(Dist.^2*ones(dim,1)));
        end
        posicion=posicion+bloque;
    end
    nonzero=find(denA>eps);
    sMap.vectors(nonzero,:)=sMap.vectors(nonzero,:)+(numA(nonzero,:)./denA(nonzero,ones(dim,1)));

    %% Tracking
    if abs(tracking)>=1,
        [qe(1,trainlen+t),qe(2,trainlen+t)]=som_quality(sMap,D);
        qe(:,trainlen+t)=real(qe(:,trainlen+t));
        trackplot([sMap.codebook sMap.values],[D F],tracking,start,t+trainlen,qe);
    end

end %for each epoch
    trainstr.end=datestr(now);
    sMap.trainhist={sMap.trainhist; trainstr};
return % function
