% Copyright (C) 2017. University of Oviedo. All rights reserved.
% Author: Ivn Machn-Gonzlez
%
% Please, if you use it, cite the following work:
% Ivn Machn-Gonzlez and Hilario Lpez-Garca, Feedforward Nonlinear Control Using Neural Gas Network, Complexity, vol. 2017, Article ID 3125073, 11 pages, 2017. doi:10.1155/2017/3125073

clear
close all
clc

disp('The plant will be simulated in open loop to obtain the training data') 
disp('The sample rate should be checked in the configuration parameters of Simulink')
sim('arm.mdl')
disp('The training data have just been obtained. They are in the variable named simout') 
fprintf(1,'\r The sample time is %.2f seconds \r',simout.TimeInfo.Increment)
fprintf(1,'The number of samples is %i  \r', size(simout.Data,1))
t=simout.Time;y=simout.Data(:,1);u=simout.Data(:,2);

figure
plot(t,y,'o-');hold on;plot(t,u,'o-k');legend('y','u');
%the sample time for modelling should be the same as considered for control in Simulink

d=1; %if d=1, the sample time is the same for data acquisition and control performance

Y=y(5*d:length(y));
Y_1=y(4*d:length(y)-1*d);
Y_2=y(3*d:length(y)-2*d);  
U=u(5*d:length(u));  
U_1=u(4*d:length(u)-1*d);  
U_2=u(3*d:length(u)-2*d);



inputs=[Y Y_1 Y_2 U_1 U_2];
n_variableaestimar=1;
numerodevariables=size(inputs,2);
inputvar={'Y_{k}','Y_{k-1}','Y_{k-2}','U_{k-1}','U_{k-2}'};

fprintf(1,'\n The target variable to be estimated is: %s \n',inputvar{n_variableaestimar})


%% Creando con la variable de entrada y aadiendo despus
D=som_data_struct(inputs,'comp_names',inputvar,'name','Arm control');
%escogiendo el nmero de columna, se define la variable de salida a estimar
D=somng_outputvar(D,n_variableaestimar); % El numero de variable n_variableaestimar  pasa a ser la variable de salida

%% Normalizacin con una funcin modificada
T=som_normalize(D,'var');

figure(15);
plot(T.data);
hold on
plot(T.values,'k-o');
legend(T.comp_names,T.value_name)

figure(16);
plot(D.data);
hold on
plot(D.values,'k-o');
legend(D.comp_names,D.value_name)


w0=som_randinit(T,'rect','msize',[5 4]);

w=ngs_batchtrain(w0,T,'trainlen',5000,'sigma',prod(w0.topol.msize)/2,'tracking',-2.010);

% trainlen = number of epochs
% sigma = initial neighborhood radius
% tracking = results for epochs. 
%             If it is equal to -2.010, it displays the results for each 10 epochs, 20 with -2.020, etc


 %we sort the number of bmu according to the center values of the target
 %variable. This is not necessary and it is only for a better visualization of the local models
  [w.values,I]=sort(w.values);
  w.vectors=w.vectors(I,:);
  w.codebook=w.codebook(I,:);



%% We save the training data and the model obtained

w_arm=w;
save w_arm w_arm
save simout simout

%% We call directly for simulation with both training and testing data
simulate