% Copyright (C) 2017. University of Oviedo. All rights reserved.
% Author: Ivn Machn-Gonzlez
%
% Please, if you use it, cite the following work:
% Ivn Machn-Gonzlez and Hilario Lpez-Garca, Feedforward Nonlinear Control Using Neural Gas Network, Complexity, vol. 2017, Article ID 3125073, 11 pages, 2017. doi:10.1155/2017/3125073
clear 
clc
close all
%We upload the previous trained model
load w_arm; w=w_arm; dim=size(w.codebook,2); 
 %We upload the data that was used for training
load simout; t=simout.Time;y=simout.Data(:,1);u=simout.Data(:,2);
d=1;% d is the time delay regarding sample time Tm for considering order of the system
Y=y(3*d:length(y));
Y_1=y(2*d:length(y)-1*d);
Y_2=y(1*d:length(y)-2*d);  
U=u(3*d:length(u));  
U_1=u(2*d:length(u)-1*d);  
U_2=u(1*d:length(u)-2*d);

 % mean and standard deviation values for target variable 'y' denormalizing
mu_values=w.value_norm.params(1);
sigma_values=w.value_norm.params(2);

%the mean values are stored in vector mu_comp, whose dimension is 1xdim
%the standard deviation values are stored in vector sigma_comp, whose dimension is 1xdim
for i=1:dim
mu_comp(i)=w.comp_norm{i}.params(1);
sigma_comp(i)=w.comp_norm{i}.params(2);
end

Yn=(Y(:)-mu_values)/sigma_values;

Yn_1  = (Y_1(:)-mu_comp(1))/sigma_comp(1); 
Yn_2 = (Y_2(:)-mu_comp(2))/sigma_comp(2); 
Un_1 = (U_1(:)-mu_comp(3))/sigma_comp(3); 
Un_2 = (U_2(:)-mu_comp(4))/sigma_comp(4); 
  
  yn_hat(1:2*d)=Yn(1:2*d); 
  BMUX(1:2*d)=NaN;
 for k=(2*d+1):length(Y)-(1*d);
     
%calculation of bmu

matrix_for_bmu=w.codebook;
bmu = som_bmus(matrix_for_bmu,[yn_hat(k-1*d) yn_hat(k-2*d) Un_1(k) Un_2(k)]);
yn_hat(k)= w.values(bmu) + w.vectors(bmu,1) * (yn_hat(k-1*d)-w.codebook(bmu,1)) + w.vectors(bmu,2) * (yn_hat(k-2*d) - w.codebook(bmu,2) ) + w.vectors(bmu,3) * (Un_1(k)-w.codebook(bmu,3))+ w.vectors(bmu,4) * (Un_2(k)-w.codebook(bmu,4)) ;
BMUX(k)=bmu; 
 end
 
 y_hat=yn_hat(:)*sigma_values+mu_values;
 
figure
plot(y_hat,'r')
hold on
plot(Y)

plot(BMUX/100,'k')
plot(U,'g')
title('Training data')

figure
 plot(Y)
 hold on
 plot(y_hat,'--r')


legend('y','y_{hat}')
title('Training data')

clear
load w_arm
 w=w_arm;

 
 disp('The plant will be simulated in open loop to obtain the test data') 
disp('The sample rate should be checked in the configuration parameters of Simulink')
 sim('arm_test.mdl')
disp('The test data have just been obtained. They are in the variable named simout') 
fprintf(1,'\r The sample time is %.2f seconds \r',simout.TimeInfo.Increment)
fprintf(1,'The number of samples is %i  \r', size(simout.Data,1))
t=simout.Time;y=simout.Data(:,1);u=simout.Data(:,2); dim=size(w.codebook,2);

% mean and standard deviation values for target variable 'y' denormalizing
mu_values=w.value_norm.params(1);
sigma_values=w.value_norm.params(2);

%the mean values are stored in vector mu_comp, whose dimension is 1xdim
%the standard deviation values are stored in vector sigma_comp, whose dimension is 1xdim
for i=1:dim
mu_comp(i)=w.comp_norm{i}.params(1);
sigma_comp(i)=w.comp_norm{i}.params(2);
end



% d is the time delay regarding sample time Tm for considering order of the system
d=1;


Y=y(5*d:length(y));
Y_1=y(4*d:length(y)-1*d);
Y_2=y(3*d:length(y)-2*d);  
U=u(5*d:length(u));  
U_1=u(4*d:length(u)-1*d);  
U_2=u(3*d:length(u)-2*d);


Yn=(Y(:)-mu_values)/sigma_values;

Yn_1  = (Y_1(:)-mu_comp(1))/sigma_comp(1); 
Yn_2 = (Y_2(:)-mu_comp(2))/sigma_comp(2); 
Un_1 = (U_1(:)-mu_comp(3))/sigma_comp(3); 
Un_2 = (U_2(:)-mu_comp(4))/sigma_comp(4); 

yn_hat_test(1:2*d)=Yn(1:2*d); 
 
 for k=(2*d+1):length(Y)-(1*d);
%calculation of bmu
matrix_for_bmu=w.codebook;
bmu = som_bmus(matrix_for_bmu,[yn_hat_test(k-(1*d)) yn_hat_test(k-(2*d))  Un_1(k) Un_2(k)  ]);
yn_hat_test(k)= w.values(bmu) + w.vectors(bmu,1) * (yn_hat_test(k-1*d)-w.codebook(bmu,1)) + w.vectors(bmu,2) * (yn_hat_test(k-2*d) - w.codebook(bmu,2) ) + w.vectors(bmu,3) * (Un_1(k)-w.codebook(bmu,3))+ w.vectors(bmu,4) * (Un_2(k)-w.codebook(bmu,4));
 end
 
 y_hat_test=yn_hat_test(:)*sigma_values+mu_values;


figure
plot(y_hat_test,'r')
hold on
plot(Y)
title('Testing')

